@@ -11,10 +11,12 @@ typedef enum {
 } unlock_state_t;
 
 typedef enum {
-    STATE_PAM_IDLE = 0,   /* no PAM interaction at the moment */
-    STATE_PAM_VERIFY = 1, /* currently verifying the password via PAM */
-    STATE_PAM_WRONG = 2   /* the password was wrong */
-} pam_state_t;
+    STATE_AUTH_IDLE = 0,         /* no authenticator interaction at the moment */
+    STATE_AUTH_VERIFY = 1,       /* currently verifying the password via authenticator */
+    STATE_AUTH_LOCK = 2,         /* currently locking the screen */
+    STATE_AUTH_WRONG = 3,        /* the password was wrong */
+    STATE_I3LOCK_LOCK_FAILED = 4 /* i3lock failed to load */
+} auth_state_t;
 
 xcb_pixmap_t draw_image(uint32_t* resolution);
 void redraw_screen(void);
